import pygame as pg

from PGUI.uielement import UIElement

def loadify(imgname):
    return pg.image.load(imgname).convert_alpha()

class Image(UIElement):
    def __init__(
            self, window,
            x: int, y: int, 
            width: int, height: int,
            image_source: str, image: pg.image = None
        ):
        super().__init__(window, x, y, width, height)

        self.image_source = image_source
        self.image = image


    def draw(self):
        super().draw()
        if self.image == None:
            image = self.loadify(self.image_source)
            self.image = pg.transform.scale(image, (self.width, self.height))

        self.window.screen.blit(self.image, (self.x, self.y))
        



import pygame as pg

from PGUI.UIElements.button import Button

class ToggleButton(Button):
    def __init__(
            self, window,
            x: int, y: int, 
            width: int, height: int, 
            text: str = "", text_size: int = 10, 
            action=None, arguments: tuple = None, 
            border_thickness: int = 3
        ) -> None:
        super().__init__(
            window, 
            x, y, 
            width, height, 
            text=text, text_size=text_size, 
            action=action, arguments=arguments, 
            border_thickness=border_thickness
        )
        self.active = False

    def on_click(self) -> None:
        super().on_click()
        self.active = not self.active

    def draw(self):
        ##If hovering or active, use colour 3
        ##If clicked, use colour 4
        ##Else use colour 2
        if self.mouse_on_element(self.window.mouse.get_pos()):
            if (self.window.mouse.get_pressed()[0]):
                button_colour = self.window.colours[3]
            else:
                button_colour = self.window.colours[2]
        else:
            button_colour = self.window.colours[1]

        if self.active:
            button_colour = self.window.colours[2]

        ##Draw border rect
        bd = self.border_thickness
        if bd > 0:
            pg.draw.rect(
                self.window.screen, self.window.colours[3],
                [self.x, self.y, self.width, self.height]
            )
        
        ##Draw button
        pg.draw.rect(
            self.window.screen, button_colour,
            [self.x+bd, self.y+bd, self.width-(bd*2), self.height-(bd*2)]
        )

        self.text.draw()

        ##Reduce text size if any is cut off
        if self.text.cut_text != "":
            self.text.font_size -= 1
            self.text.pg_font = pg.font.Font(
                self.text.font, self.text.font_size
            )


from typing import List
import pygame as pg

class UIElement:
    """All UI elements must inherit from this class. Gives basic methods and
        attributes for UI use.
    """
    def __init__(
                self, 
                window,
                x: int, y: int, 
                width: int, height: int, 
                border_thickness: int = 3
            ) -> None:

        ##Drawing details.
        from PGUI.window import Window
        self.window: Window = window
        "The parent window of the UI element."
        self.x = x
        "The X screen position of the element."
        self.y = y
        "The Y screen position of the element."
        self.width = width
        "The width of the element in pixels."
        self.height = height
        "The height of the element in pixels."
        self.border_thickness = border_thickness
        "How thick the border of the element should be."
        self.drawing = True
        """Controls whether or not the element is drawn."""

        ##Tooltip stuff.
        self.use_tooltip = False
        """Whether to display a tooltip for this element or not."""
        self.tooltip_title = ""
        """The title for the tooltip class to draw."""
        self.tooltip_text = ""
        """The text for the tooltip class to display."""

        ##Child stuff.
        self.is_child = False
        """True if element is a child of another element."""
        self.children: List[UIElement] = []
        """Child elements of element."""

        ##If parent is another UI element, add self as child and get window.
        if isinstance(self.window, UIElement):
            self.is_child = True
            self.window.add_child(self)
            self.window = self.window.window
            return
        else:
            ##Add element to window.
            self.window.add_ui_element(self)



    #region mouse
    def mouse_on_element(self, mouse) -> bool:
        """Returns true if mouse is within element bounding box"""
        bd = self.border_thickness
        return (
            (self.x + bd <= mouse[0] <= self.x + self.width - (bd*2)) and
            (self.y + bd <= mouse[1] <= self.y + self.height - (bd*2))
        ) 


    def on_click(self) -> None:
        """Called when element is clicked."""
        ##Does nothing by default.
        pass


    def clicked(self, button=0) -> bool:
        return (self.mouse_on_element(self.window.mouse.get_pos()) 
                and any([event.type == pg.MOUSEBUTTONDOWN for event in self.window.events])
                and self.window.mouse.get_pressed()[button]
            )


    def mouse_held(self, button=0) -> bool:
        """Returns true if element is clicked and held."""
        return (self.mouse_on_element(self.window.mouse.get_pos()) 
                and self.window.mouse.get_pressed()[button])
    #endregion



    def add_child(self, child):
        """
        Add a child UI Element to the element.
        Note: if this is not called to add a child element it may break
        automated element spacing within windows."""
        self.children.append(child)
        child.is_child = True


    def update(self) -> None:
        """Called once every frame."""
        if not self.window.running:
            return

        if self.clicked():
            self.on_click()

        for child in self.children:
            child.update()


    def draw(self) -> None:
        """Called once every frame and draws the element."""
        if not self.drawing:
            return

        for child in self.children:
            child.draw()


    def update_colours(self) -> None:
        if self.children:
            for child in self.children:
                child.update_colours()


    def set_x(self, new_x):
        self.x = new_x
        if self.children:
            for child in self.children:
                child.set_x(new_x)


    def set_y(self, new_y):
        self.y = new_y
        if self.children:
            for child in self.children:
                child.set_y(new_y)


    def set_width(self, new_width):
        self.width = new_width
        if self.children:
            for child in self.children:
                child.set_width(new_width)
    

    def set_height(self, new_height):
        self.height = new_height
        if self.children:
            for child in self.children:
                child.set_height(new_height)


import pygame as pg 

from PGUI.uielement import UIElement
from PGUI.UIElements.text import RawText, WrappedText

class Tooltip(UIElement):
    """Adds a tooltip that follows the mouse and displays information about
    any UI elements the user hovers over. UIElement.use_tooltip must be True
    to use this feature and it is False by default.
    - UIElement.tooltip_title sets the title for the tooltip
    - UIElement.tooltip_text sets the body text for the tooltip

    The tooltip will automatically size itself to fit the text it's 
    displaying.
    """

    def __init__(
            self, window,
            border_thickness: int = 1,
        ) -> None:
        super().__init__(window, 0, 0, 200, 50, border_thickness)
        self.drawing = False

        self.title = RawText(
            self,
            0, 0, "", 
            (window.font, 15), 2
        )

        self.text = WrappedText(
            self,
            0, 0, 400, 10000, "", 
            (window.font, 10), 2
        )

    def get_element_hovering(self) -> UIElement:
        """Get highest-layer element mouse is currently hovering over with
        use_tooltip enabled."""
        elements_hovering = [
            element for element in self.window.ui_elements
            if element.mouse_on_element(self.window.mouse.get_pos())
            and element.use_tooltip
        ]
        if elements_hovering == []:
            return None
        else:
            ##Return last element in list (highest render priority)
            return elements_hovering[-0]

    def update(self) -> None:
        """Update position to mouse position and set title and text to that of
        the UIElement of the highest position in window.ui_elements."""
        mouse_pos = self.window.mouse.get_pos()

        ##Get highest element.
        element = self.get_element_hovering()
        ##Return if not hovering over tooltip-enabled element.
        if element == None:
            self.drawing = False
            return

        self.drawing = True
        self.title.text = element.tooltip_title

        ##Fit tooltip to title size if body text not present.
        if element.tooltip_text == "":
            self.tooltip_text = ""
            self.height = self.title.get_size()[1] + 4
            self.width = self.title.get_size()[0]
        else:
            ##set width to widest text.
            self.width = (
                max(
                    self.title.get_size()[0],
                    self.text.text_width
                ) + 2
            )
            self.text.text = element.tooltip_text
            ##Set height to sum of title and text height.
            self.height = (
                self.title.get_size()[1] 
                + self.text.text_height 
                + 8
                )

        ##Put tooltip left of mouse if rendering it right would go off-screen.
        if self.width + mouse_pos[0] > self.window.screen_size[0]:
            self.x = mouse_pos[0]-self.width-5
        else:
            self.x = mouse_pos[0]

        ##Put tooltip above mouse if rendering it below would go off-screen.
        if self.height + mouse_pos[1] > self.window.screen_size[1]:
            self.y = mouse_pos[1]-self.height-5
        else:
            self.y = mouse_pos[1]+5

        self.title.x, self.title.y = self.x+2, self.y+2
        self.text.x, self.text.y = self.x+2, self.y+22

    def draw(self) -> None:
        """Draw tooltip to parent window."""
        if not self.drawing:
            return None

        super().draw()
        draw = pg.draw
        window = self.window
        bd = self.border_thickness

        ##Draw border rect.
        draw.rect(
            window.screen, window.colours[3], 
            [self.x-bd, self.y-bd, self.width+2*bd, self.height+2*bd]
        )
        ##Draw inner rect.
        draw.rect(
            window.screen, window.colours[0], 
            [self.x, self.y, self.width, self.height]
        )
        ##Draw text
        self.title.draw()
        self.text.draw()


"""Adds UI elements based on pygame."""

from os.path import dirname, basename, isfile, join
import glob
modules = glob.glob(join(dirname(__file__), "*.py"))
__all__ = [basename(f)[:-3] for f in modules if isfile(f) and not f.endswith('__init__.py')]

from PGUI.UIElements.button       import Button
from PGUI.UIElements.colourpicker import ColourPicker
from PGUI.UIElements.dropdown     import DropDown
from PGUI.UIElements.fpsdisplay   import FPSDisplay
from PGUI.UIElements.image        import Image
from PGUI.UIElements.inputfield   import InputField
from PGUI.UIElements.panel        import Panel
from PGUI.UIElements.slider       import Slider
from PGUI.UIElements.text         import RawText, WrappedText
from PGUI.UIElements.titlescreen  import TitleScreen
from PGUI.UIElements.togglebutton import ToggleButton
from PGUI.UIElements.tooltip      import Tooltip


import pygame as pg

from PGUI.uielement import UIElement

class Panel(UIElement):

    def draw(self):
        super().draw()
        bd = self.border_thickness
        ##Border rect.
        pg.draw.rect(
            self.window.screen, self.window.colours[3], 
            [self.x, self.y, self.width, self.height]
        )
        ##Colour rect.
        pg.draw.rect(
            self.window.screen, self.window.colours[0], 
            [self.x+bd, self.y+bd, self.width-(bd*2), self.height-(bd*2)]
        )


from PGUI.UIElements.text import RawText

class FPSDisplay(RawText):
    def __init__(
            self, window, 
            x: int, y: int, 
            font, colour: int, 
            clock,
        ) -> None:
        super().__init__(
            window, 
            x, y,
            "", font, 
            colour,
        )
        self.clock = clock

    def update(self) -> None:
        fps = round(self.clock.get_fps())
        self.text = f"FPS: {fps}"


def space_elements(
            elements: "list[int]",
            size: int,
            spacing: int,
            padding: int,
            position: int = 0
        ) -> tuple:
    """ Set size and position of elements to space them correctly in a
        given distance."""

    """Calculate each element's width."""
    ##Remove padding from either end of distance.
    size -= padding*2
    ##Calculate total element space without padding or spacing.
    element_space = size - spacing * (len(elements)-1)
    ##Divide total space by sum of the ratio between elements.
    multiplier = element_space/sum(elements)
    ##Multiply by ratio to get each element width.
    element_widths = [round(e*multiplier) for e in elements]

    """Calculate each element's position."""
    element_positions = []
    ##Calculate position and add to list.
    for i in range(len(elements)):
        element_positions.append(round(
            spacing * i
            + sum(element_widths[:i]) ##Add previous element widths.
            + padding
            + position
        ))
    
    return element_positions, element_widths


"""Adds a window, innit?"""

import pygame as pg
from PGUI.uielement import UIElement

class Window:
    """Creates a window using pygame.
    - UIElements can be added to the window, and their update() and draw()
      methods will be called every frame.
    """
    #region meta
    def __init_subclass__(cls, window_id: str, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        ##Each subclass of Window must have a window_id.
        cls.window_id = window_id

    def __str__(self) -> str:
        screen_size = f"(X){self.screen_size[0]}px (Y){self.screen_size[1]}px"
        return (f"Window '{self.window_id}': title = {self.title}, size = {screen_size}")
    #endregion

    def __init__(
            self, 
            parent, 
            title: str, 
            width: int, 
            height: int,
        ) -> None:

        ##Display stuff.
        self.screen_size: tuple = (width, height)
        """The width and height of the screen in pixels."""
        self.title: str = title
        """The title to be displayed at the top of the window."""
        from PGUI.program import Program
        self.parent: Program = parent
        """Window's parent program."""
        self.colours = self.parent.colours
        """List of colours used by the window."""

        ##Pygame stuff.
        self.running: bool = False
        self.fps: int = 60
        """Number of frames per second to draw the window at."""
        self.clock = pg.time.Clock()
        """Pygame clock to control FPS."""
        self.mouse = pg.mouse
        """Reference to the pygame mouse."""
        self.events = []
        """List of pygame events, updated every frame."""

        ##Font.
        pg.font.init()
        self.font: str = "Assets/Fonts/GravityRegular5.ttf"
        """The text font to be used throughout the window."""

        ##List of UI elements.
        self.ui_elements: list[UIElement] = []
        """List of UI elements that the window draws and updates every 
        frame."""

        ##Add self to controller.
        self.parent.add_window(self, self.window_id)


    def add_ui_element(self, ui_element: UIElement) -> None:
        """Adds a UI element to the window."""
        self.ui_elements.append(ui_element)
        ui_element.window = self


    def add_ui_elements(self, ui_elements: list) -> None:
        """Adds multiple UI elements to the window."""
        for ui_element in ui_elements:
            self.ui_elements.append(ui_element)
            ui_element.window = self


    def remove_ui_element(self, ui_element: UIElement) -> None:
        """Removes a UI element from the window."""
        self.ui_elements.remove(ui_element)


    def update(self) -> None:
        """Called every frame"""
        ##Get events. 
        ##In try except because program crashes on close otherwise.
        try:
            self.events = pg.event.get()
        except pg.error:
            self.end()

        ##Close window if close button clicked or parent stops running.
        if (
            any(event.type == pg.QUIT for event in self.events) 
            or not self.parent.running
           ):
            self.end()
            return
        
        ##Update the window's UI elements.
        for ui_element in self.ui_elements:
            ui_element.update()


    def draw(self) -> None:
        """Called every frame and draws the window."""
        if not self.parent.running:
            return

        ##Fills the screen. Drawing code run before this will not display.
        self.screen.fill(self.colours[0])

        ##Draws each of the UI elements in the window.
        for ui_element in self.ui_elements:
            ui_element.draw()

        ##Displays what's been drawn.
        pg.display.flip()

    def run(self) -> None:
        """
        Starts the window running. This is called once when the window is
        first opened."""
        self.running = True
        self.setup()

        while self.running:
            self._run()
            self.clock.tick(self.parent.fps)

    def setup(self) -> None:
        """
        Sets the window up. If this isn't called before the window is
        opened it will retain the old window's size and title."""
        self.screen = pg.display.set_mode(self.screen_size)
        pg.display.set_caption(self.title)
        
    def _run(self) -> None:
        """
        Runs the window. This function contains anything that needs to
        happen every frame."""     
        self.update()
        self.draw()


    def end(self) -> None:
        """
        Closes the window. Any code that needs to happen before the window
        closes goes here."""
        self.running = False


    def open_window(self, window_id):
        """
        Opens another window and then runs own setup method. This prevents
        window A from retaining window B's size when B is closed."""
        window = self.parent.get_window(window_id)
        window.run()
        self.setup()


    def update_colours(self, colours: list):
        self.colours = colours
        for element in self.ui_elements:
            element.update_colours()



from PGUI.spacing import space_elements
import PGUI.UIElements as elements

class EditingDisplay(elements.Panel):
    def __init__(
            self, window, 
            x: int, y: int, 
            width: int, height: int, 
            border_thickness: int = 3
        ) -> None:
        super().__init__(
            window, 
            x, y, width, height, 
            border_thickness=border_thickness
        )

        elements.RawText(
            self,
            605, 175,
            "Editing Map:",
            (self.window.font, 10), 3
        )

        self.map_name_text = elements.RawText(
            self,
            605, 190,
            "",
            (self.window.font, 10), 2
        )

        elements.RawText(
            self,
            605, 205,
            "Selected Tile:",
            (self.window.font, 10), 3
        )

        self.tile_name_text = elements.RawText(
            self,
            605, 220,
            "",
            (self.window.font, 10), 2
        )

        elements.RawText(
            self,
            605, 235,
            "Description:",
            (self.window.font, 10), 3
        )

        self.tile_desc_text = elements.WrappedText(
            self,
            605, 250,
            165, 200,
            "",
            (self.window.font, 10), 2, False
        )

        y_positions, heights = space_elements(
            [1, 1, 1, 1, 1, 10],
            self.height,
            5, 5, self.y
        )

        for index, element in enumerate(self.children):
            element.set_y(y_positions[index])
            element.set_height(heights[index])

    def update(self) -> None:
        super().update()
        self.map_name_text.text = self.window.editor_grid.map_name

        tile = self.window.parent.selected_tile
        if tile:
            self.tile_name_text.text = tile.tile_name
            self.tile_desc_text.text = tile.tile_description


    def draw(self):
        super().draw()
        for child in self.children:
            child.draw()



import pygame as pg

from PGUI.uielement import UIElement

class RawText(UIElement):
    """Displays text in a given font and size.
    Note: Does not account for width or height constraints.
    
    Parameters
    ----------
    - All `UI_Element` parameters minus width and height.
    - `text: str` - The text to be displayed.
    - `font: tuple(font_path: str, font_size: int)` - The font in which to
      display the text.
    - `colour: tuple or int` - The colour to draw the text in.

    Methods
    -------
    - All `UI_Element` methods.
    - `get_size() -> tuple(int, int)` - Get size of the text.
    """
    def __init__(
            self, window,
            x: int, y: int, 
            text: str, font, colour,
        ) -> None:

        super().__init__(window, x, y, 0, 0)

        self.text = text
        self.font: pg.font.Font = pg.font.Font(*font)

        ##If colour is int, use it as index of window's colours.
        if isinstance(colour, int):
            self.colour = self.window.colours[colour]
            self.colour_index = colour
        ##If colour is list or tuple, use values to make colour.
        elif isinstance(colour, (tuple, list)):
            self.colour = colour
            self.colour_index = None
        else:
            raise TypeError("Invalid colour argument.")


    def draw(self) -> None:
        """Draw text to window."""
        text = self.font.render(self.text, False, self.colour)
        self.window.screen.blit(text, (self.x, self.y))


    def get_size(self) -> tuple:
        """Return size of text."""
        return self.font.size(self.text)


    def update_colours(self) -> None:
        if self.colour_index:
            self.colour = self.window.colours[self.colour_index]






class WrappedText(RawText):
    """Wraps text to fit within width and height constraints.
    """
    def __init__(
            self, window,
            x: int, y: int, 
            width: int, height: int,
            text: str, font, colour: tuple, centred: bool = False,
        ) -> None:
        super().__init__(window, x, y, text, font, colour)

        self.width = width
        self.height = height

        self.font = font[0]
        self.font_size = font[1]
        self.pg_font = pg.font.Font(*font)
        self.centred = centred
        self.text_width = 0
        self.text_height = 0

        ##Any text that doesn't fit
        self.cut_text = ""


    def draw(self) -> None:
        """Draws the text wrapped in a box."""
        ##str() called so that changes to text don't change self.text
        text = str(self.text)

        ##Get the height of the tallest letter in the font
        font_height = self.pg_font.size("T")[1]
        line_spacing = 3
        
        display = []
        i = 0
        y = int(self.y)
        width = 0
        new_line = False

        while text:

            new_line = False

            ##Stops loop if out of y space.
            if y + font_height > y + self.height:
                break

            ##Increments i until it runs out of x space.
            while self.pg_font.size(text[:i])[0] < self.width and i < len(text):
                i += 1
                if text[i-1] == "\n":
                    new_line = True
                    break

            ##Finds the last space if there is still text to display.
            if i < len(text) and new_line == False:
                i = text.rfind(" ", 0, i) + 1

            if i != 0:
                if (text[:i][-1] == " " or text[:i][-1] == "\n"):
                    display.append(
                        self.pg_font.render(text[:i-1], False, self.colour)
                    )
                else:
                    display.append(
                        self.pg_font.render(text[:i], False, self.colour)
                    )

            new_line = False
            text = text[i:]
            if i == 0:
                break

        total_height = len(display) * (font_height + line_spacing)

        for image in display:
            if self.centred:
                self.window.screen.blit(
                    image, 
                    (
                        self.x+(self.width/2)-image.get_rect().width/2, 
                        y+(self.height/2)-(total_height/2))
                    )
            else:
                width = max(width, image.get_rect()[2])
                self.window.screen.blit(image, (self.x, y))
                
            y += font_height+line_spacing

        ##Return actual width and height of text
        self.text_height = y - self.y
        self.text_width = width

        if text:
            self.cut_text = text
        else:
            self.cut_text = ""


import pygame as pg
import os.path
import json

from PGUI.spacing import space_elements
from PGUI.uielement import UIElement
from PGUI.UIElements.slider import Slider
from PGUI.UIElements.button import Button
from PGUI.UIElements.text import RawText
from PGUI.UIElements.panel import Panel

class ColourPicker(UIElement):
    """Creates 3 sliders for each colour in the parent window palette.
    - Users can change each RGB value and see the colours update in the
      preview boxes next to them in real time.

    Also adds 3 buttons:
    - "Apply" button will save the selected colours to the parent window and
      it's parent controller (if present).
    - "Reset" button sets the sliders back to their last saved state.
    - "Defaults" button sets the values back to their hard-coded defaults.
    """
    def __init__(
            self, window,
            x: int, y: int, 
            width: int, height: int,
            border_thickness: int = 3
        ) -> None:
        super().__init__(window, x, y, width, height, border_thickness)

        self.colours = list(self.window.colours)
        self.palette_directory = "Assets/palettes/"
        self.palettes = self.load_palettes()
        self.palette_index = -1
        self.ugh = False

        ##Child UI elements
        self.colour_sliders = []
        self.channel_text = []
        self.colour_text = []
        self.buttons = []

        ##Background panel
        Panel(
            self,
            x, y, 
            width, height-border_thickness, 
            border_thickness
        )

        element_ratios = []
        for colour in self.window.colours:
            element_ratios.append(1)
            for channel in colour:
                element_ratios.append(1)
        ##Buttons are twice the height of each slider.
        element_ratios.append(2)

        y_positions, heights = space_elements(
            elements=element_ratios,
            size=self.height-(border_thickness*2),
            spacing=5,
            padding=5,
            position=self.y+border_thickness
        )

        self.slider_height = heights[2]
        self.slider_spacing = (
            y_positions[2]
            - y_positions[1]
            - self.slider_height
        )

        self.preview_size = (heights[2]*3)+(self.slider_spacing*2)
        self.preview_positions = []

        self.slider_width = (
            self.width
            - self.preview_size * 2
            - 15
        )

        ##Colour sliders, text and previews
        i = 0
        for col_i, colour in enumerate(self.window.colours):

            ##Add text for each colour
            window = self.window
            text = RawText(
                self,
                x=self.x+7, y=y_positions[i],
                text=f"Colour {col_i+1}:",
                font=(window.font, heights[i]), 
                colour=3
            )
            self.colour_text.append(text)
            i += 1

            self.preview_positions.append(y_positions[i])
            for chan_i, channel in enumerate(colour):
                ##Change slider width based on element width
                slider_width = self.slider_width

                ##Add slider for each colour channel.
                slider = Slider(
                    self,
                    self.x+7, y_positions[i], slider_width, heights[i],
                    0, 255, channel
                )
                self.colour_sliders.append(slider)

                ##Add value text for each colour channel
                text = RawText(
                    self,
                    self.x+5+slider_width+5, y_positions[i], 
                    "Value", (self.window.font, heights[i]),
                    colour=3
                )
                self.channel_text.append(text)
                i += 1

        #region buttons
        ##Align and justify buttons.
        button_positions, button_widths = space_elements(
            elements=[2, 2, 3, 2],
            size=self.width,
            spacing=5,
            padding=7,
            position=self.x
        )

        ##Apply changes button
        self.apply_button = Button(
            self,
            button_positions[0], y_positions[i], button_widths[0], heights[i],
            "Apply", 13, action=self.apply
        )
        self.buttons.append(self.apply_button)

        ##Revert changes button
        self.reset_button = Button(
            self,
            button_positions[1], y_positions[i], button_widths[1], heights[i],
            "Reset", 13, action=self.reset
        )
        self.buttons.append(self.reset_button)

        ##Reset to defaults button
        self.default_button = Button(
            self,
            button_positions[2], y_positions[i], button_widths[2], heights[i],
            "Defaults", 13, action=self.defaults
        )
        self.buttons.append(self.default_button)

        ##Load button
        self.load_button = Button(
            self,
            button_positions[3], y_positions[i], button_widths[3], heights[i],
            "Load", 13, action=self.switch_palette
        )
        self.buttons.append(self.load_button)
        #endregion

        self.config = read_config("config.json")
        self.load_colours()

    ##Called every frame
    def update(self) -> None:
        super().update()

        ##Update colours for previews
        for i in range(0, len(self.colours)):
            self.colours[i] = ( self.colour_sliders[(i*3)+0].get_value(),
                                self.colour_sliders[(i*3)+1].get_value(),
                                self.colour_sliders[(i*3)+2].get_value())

        ##Check buttons clicked.
        for button in self.buttons:
            if button.clicked():
                button.action()

    def draw(self) -> None:
        super().draw()
        preview_size = self.preview_size

        ##Draw preview square of each colour
        i = 0
        for colour in self.colours:
            for channel in range(0, len(colour)):
                self.channel_text[(i*3)+channel].text = str(colour[channel])

            ##Base colour preview rect.
            preview_rect = [
                self.x+self.width-self.preview_size-self.border_thickness-5,
                self.preview_positions[i],
                preview_size, preview_size
            ]

            ##Background border rect
            pg.draw.rect(self.window.screen, (0,0,0), preview_rect)

            ##Coloured preview
            pg.draw.rect(self.window.screen, self.colours[i],
                         [preview_rect[0]+self.border_thickness,
                          preview_rect[1]+self.border_thickness,
                          preview_rect[2]-(self.border_thickness*2),
                          preview_rect[3]-(self.border_thickness*2)])
            i += 1


    ##Apply selected colours to parent window
    def apply(self, save=True) -> None:
        self.window.colours = list(self.colours)

        all_text = self.channel_text + self.colour_text
        for text in all_text:
            text.colour = self.window.colours[3]
        
        if self.window.parent.colours:
            self.window.parent.set_colours(list(self.colours))

        if save:

            username = self.window.parent.user
            if not username:
                return

            ##Get path to user options file.
            user_options_path = f"{self.config.user_options_folder}{username}.json"

            ##If file exists, get contents.
            if os.path.exists(user_options_path):
                with open(user_options_path, "r") as file:
                    options = json.load(file)
                    options["colours"] = self.colours
                    file.close()
            else:
                ##Make new options if file does not exist.
                options = {"colours": self.colours}


            with open(user_options_path, "w") as file:
                json.dump(options, file)
                file.close()

            self.load_button.text.text = "user"


    ##Revert colours to last selection
    def reset(self) -> None:
        self.colours = list(self.window.colours)
        self.update_sliders()


    ##Revert selection to default
    def defaults(self) -> None:
        self.colours = self.config.default_colours
        self.update_sliders()
        self.apply(False)


    ##Update slider values to match colours
    def update_sliders(self) -> None:
        i = 0
        for colour in self.colours:
            for channel in colour:
                self.colour_sliders[i].set_value(channel)
                i += 1


    def load_colours(self) -> None:
        username = self.window.parent.user
        config = self.config
        default_colours = config.default_colours
        if not username:
            self.colours = default_colours
            self.update_sliders()
            self.apply(False)
            return

        user_options_path = f"{config.user_options_folder}{username}.json"
        
        ##Create palette file if one does not already exist.
        if not os.path.isfile(user_options_path):
            self.defaults()
            with open(user_options_path, "w+") as file:
                json.dump({"colours":default_colours}, file)
                file.close()
        else:
            with open(user_options_path, "r") as file:
                self.colours = json.load(file)["colours"]
                file.close
            self.update_sliders()
            self.apply(False)


    def load_palettes(self):
        palettes = []
        for palette in [i for i in os.listdir(self.palette_directory) if i[-5:] == ".json"]:
            with open(self.palette_directory+palette) as file:
                contents = file.read()
                palettes.append((palette[:-5], json.loads(contents)))
        return palettes

    def switch_palette(self):
        ##I have no idea why, but clicking the switch palette button
        ##calls this function twice, so the bool ugh stops it from
        ##actually running twice.
        self.ugh = not self.ugh
        if self.ugh:
            self.palette_index = (self.palette_index + 1) % len(self.palettes)
            self.colours = self.palettes[self.palette_index][1]
            self.load_button.text.text = self.palettes[self.palette_index][0]
            self.update_sliders()
            self.apply(False)



from PGUI.window import Window
import pygame as pg

class Program():
    """
    A class that stores and controls windows.

    Window navigation is designed to work like the internet, with each
    window having a unique identifier, used to get reference to it using
    the program's `get_window` function.
    """


    _windows = {}
    """
    Window registry. Each subclass of window has a unique `window_id` 
    string, used as the key to this dictionary, with the window 
    instance being the value.
    """

    def __init__(self,
            default_window = "", ## Window ID of the default window.
            fps: int = 60,       ## Frames per second to draw windows at.
        ) -> None:

        ##The number of frames per second at which to draw windows.
        self.fps = fps
        """Number of frames per second to be drawn."""
        self.default_window = default_window
        """Window ID of the window to be opened first."""

        ##Tracks if the program is running or not.
        self.running = False
    

    def add_window(self, window: Window, window_id: str) -> None:
        """
        Add a window to the program, set itself as the parent and apply 
        variables such as fps.

        Note: If the window's id is `_load_`, it will be run as soon as
        the program starts, before the default window.
        """
        ##Set window attributes
        window.parent = self
        window.fps = self.fps

        ##Add window to windows dictionary.
        self._windows[window_id] = window

        if window_id == "_load_":
            window.run()


    def run(self) -> None:
        """Start running the program's default window."""
        self.running = True

        pg.init()

        ##If loading window exists, run it first.
        if "_load_" in self._windows.keys():
            self.get_window("_load_").run()

        ##Run default window.
        self.get_window(self.default_window).run()

        ##End the program when all windows have ended.
        self.end()


    def end(self) -> None:
        """Quit the program."""
        self.running = False
        for window in self._windows.values():
            window.end()
        pg.quit()
        print("\nProgram ended.")


    def get_window(self, window_id: str) -> Window:
        "Return a window from it's window_id."
        return self._windows[window_id]


    def set_fps(self, fps: int) -> None:
        """Set the framerate of the program."""
        self.fps = int(fps)


from PGUI.UIElements.panel import Panel
from PGUI.UIElements.text import WrappedText

class TitleScreen(Panel):
    def __init__(
            self, window,
            x: int, y: int, 
            width: int, height: int, 
            text, font,
            border_thickness: int = 3
        ) -> None:
        super().__init__(
            window, 
            x, y, 
            width, height, 
            border_thickness=border_thickness
        )
        self.text = text
        self.font = font
    
    def draw(self) -> None:
        super().draw()
        WrappedText(
            self.x, self.y, self.width, self.height, self.text, self.font,
            self.window.colours[3], window=self.window, centred=True
        ).draw()


from typing import Tuple
import pygame as pg

from PGUI.uielement import UIElement
from PGUI.UIElements.text import WrappedText

class Button(UIElement):
    """
    Creates a button that the user can press to call an action.

    Parameters
    ----------
    - All `UIElement` parameters.
    - `text: str` - The text to display on the button.
    - `text_size: int` - The size (in pixels) to display the text.
    - `action: method` - The method to call when the button is clicked.
    - `arguments: tuple` - Arguments to pass to the method or function.

    Methods
    -------
    - All `UIElement` methods.
    - `update() -> None` - Called by the parent window every frame to update 
      the button.
    - `draw() -> None` - Called by the parent every frame to draw the button.
    """

    def __init__(
            self, window,
            x: int, y: int, 
            width: int, height: int,
            text: str = "", text_size: int = 10, 
            action = None, arguments: tuple = None,
            border_thickness: int = 3
        ) -> None:

        ##Initialise the UIElement.
        super().__init__(window, x, y, width, height, border_thickness)

        self.text = WrappedText(
            self,
            self.x+5, self.y+5,
            self.width-10, self.height-10,
            text, (self.window.font, text_size), 
            self.window.colours[0],
            True
        )
        """Text to display on button."""

        self.action = action
        """Method or function for the button to call on click."""
        self.arguments = arguments
        """Arguments to be passed to button function or method."""


    def on_click(self) -> None:
        """Calls the button's argument(s)."""
        if self.action != None:
            if self.arguments == None: 
                self.action()
            elif isinstance(self.arguments, Tuple):
                self.action(*self.arguments)
            else:
                self.action(self.arguments)


    def draw(self):
        """Draw the button to the screen."""

        mouse = self.window.mouse
        ##If hovering, use colour 3.
        if self.mouse_on_element(mouse.get_pos()):
            ##If clicked, use colour 3.
            colour_index = 3 if mouse.get_pressed()[0] else 2
        else:
            ##If not hovering use colour 2.
            colour_index = 1

        button_colour = self.window.colours[colour_index]

        ##Draw border rect.
        bd = self.border_thickness
        if bd > 0:
            pg.draw.rect(
                self.window.screen, 
                self.window.colours[3],
                [self.x, self.y, self.width, self.height]
            )

        ##Draw button.
        pg.draw.rect(
            self.window.screen, 
            button_colour,
            [self.x+bd, self.y+bd, self.width-(bd*2), self.height-(bd*2)]
        )

        self.text.draw()

        ##Reduce text size if any is cut off
        if self.text.cut_text != "":
            self.text.font_size -= 1
            self.text.pg_font = pg.font.Font(self.text.font, self.text.font_size)


    def update_colours(self) -> None:
        super().update_colours()
        self.text.colour = self.window.colours[0]


import pygame as pg

from PGUI.uielement import UIElement

class Slider(UIElement):
    """Creates a slider that can be dragged by the mouse.
    """
    def __init__(
            self, window,
            x: int, y: int, 
            width: int, height: int,
            min_value: int, max_value: int, default_value: int = 0,
            border_thickness: int = 3
        ) -> None:
        super().__init__(window, x, y, width, height, border_thickness)

        self.min_value = min_value
        self.max_value = max_value
        self.set_value(default_value)

        self.being_dragged = False


    def set_value(self, value) -> None:
        """Sets the position of the slider from a value within its 
        min_value and max_value.
        """
        self.pos_value = ((value-self.min_value)/self.max_value)*self.width


    def get_value(self) -> int:
        """Returns the value of the slider."""

        return min(
            int((self.pos_value/self.width)*self.max_value)+self.min_value, 
            self.max_value
        )


    def update(self) -> None:
        """Called every frame."""
        super().update()

        mouse = self.window.mouse
        mouse_x = mouse.get_pos()[0]
        bd = self.border_thickness

        ##Drag starts when element is clicked on.
        if self.clicked(0):
            self.being_dragged = True

        ##Drag continues if mouse within element x and left click held.
        self.being_dragged = (
            self.being_dragged
            and self.x+bd <= mouse_x <= self.x+self.width-bd
            and mouse.get_pressed()[0]
        )
        
        ##Update slider position if being dragged.
        if self.being_dragged:
            self.pos_value = mouse_x-self.x+bd



    def draw(self) -> None:
        """Draws the slider to the parent window."""
        super().draw()

        ##Useful little things.
        bd = self.border_thickness
        colours = self.window.colours
        screen = self.window.screen

        ##Draw slider border.
        if bd > 0:
            pg.draw.rect(
                screen, colours[3],
                [self.x, self.y, self.width, self.height]
            )

        ##Draw unfilled area of slider.
        pg.draw.rect(
            screen, colours[1],
            [self.x+bd, self.y+bd, self.width-(bd*2), self.height-(bd*2)]
        )

        ##Draw filled area of slider.
        pg.draw.rect(
            screen, colours[2],
            [self.x+bd, self.y+bd, self.pos_value-(bd*2), self.height-(bd*2)]
        )


import pygame as pg
import string

from pygame.constants import KEYDOWN

from PGUI.UIElements.togglebutton import ToggleButton

class InputField(ToggleButton):
    #region perameters
    def __init__(
            self, window,
            x: int, y: int, 
            width: int, height: int, 
            text: str = "", 
            numbers_only: bool = False, 
            censored: bool = False,
            text_size: int = 10, 
            border_thickness: int = 3,
        ):
        super().__init__(
            window,
            x, y, width, height, 
            text=text, 
            text_size=text_size, 
            action=None, arguments=None, 
            border_thickness=border_thickness, 
        )
    #endregion

        self.active = False
        self.text_value = text
        self.censored = censored

        ##Set allowed characters based on numbers_only.
        if numbers_only:
            self.ALLOWED_CHARS = string.digits
        else:
            self.ALLOWED_CHARS = (
                string.digits
                + string.ascii_letters 
                + string.punctuation
                ##string.whitespace isn't used because it allows return.
                + " "
            )

    def update(self):
        super().update()
        ##Show visible text as "*" symbols if input box is censored.
        if self.censored:
            self.text.text = "*" * len(self.text_value)
        else:
            self.text.text = self.text_value

        ##If mouse is clicked, set active if it is on element.
        if self.window.mouse.get_pressed()[0]:
            self.active = self.mouse_on_element(self.window.mouse.get_pos())

        ##Check for key input when active.
        if self.active:
            key_events = [i for i in self.window.events if i.type == pg.KEYDOWN]
            for key_down in key_events:
                if key_down.key == pg.K_BACKSPACE:
                    self.text_value = self.text_value[:-1]
                    continue
                elif key_down.unicode in self.ALLOWED_CHARS:
                    self.text_value += key_down.unicode

    def get_value(self):
        """Return input field value."""
        return self.text_value


import pygame as pg

from PGUI.UIElements.button import Button

class DropDown(Button):
    def __init__(
        self, window, 
        x: int, y: int, 
        width: int, height: int, 
        text: str, items: tuple, dropdown_action=None,
        default_item = None,
        text_size=10, border_thickness: int = 3
    ) -> None:
        super().__init__(
            window, x, y, width, height, text, text_size,
            border_thickness=border_thickness
        )
        self.items = items
        self.buttons = []
        self.selected_item = None
        self.dropdown_text = text
        self.dropdown_action = dropdown_action
        self.open = False
        self.text_size = text_size
        i = self.y
        for item in self.items:
            button = Button(
                    self,
                    self.x+2, self.height+i+2, self.width-4, self.height-4, 
                    str(item), text_size, self.update_selected, (item), 
                    1
                )
            self.buttons.append(button)
            i += self.height
        
        self.update_selected(default_item)

    def update(self) -> None:
        super().update()
        if self.clicked():
            self.open = not self.open
        if self.open:
            for button in self.buttons:
                button.update()

    def draw(self) -> None:
        super().draw()
        if self.open:
            for button in self.buttons:
                button.draw()

    def update_selected(self, item) -> None:
        self.selected_item = item
        self.open = False
        self.text.text = f"{self.dropdown_text}: {item}"
        if self.dropdown_action != None:
            self.dropdown_action(item)